/*
 * MUST (Marmot Umpire Scalable Tool)
 * Copyright (C) ZIH, Technische Universitaet Dresden, Federal Republic of Germany, 2011-2012
 * Copyright (C) Lawrence Livermore National Laboratories, United States of America, 2011-2012
 *
 * See the file LICENSE.txt in the package base directory for details
 */

/**
 * @file LocationWrapper.h
 * 		Header for the wrapper operation to gather locations.
 *
 * @author Tobias Hilbrich
 * @date 04.01.2011
 */

#include <stdint.h>
#include <map>
#include <string>
#include <mpi.h>
#include <pnmpi.h>
#include <assert.h>

#include "mustConfig.h"

#include "BaseIds.h"
#include "MustDefines.h"

#ifdef USE_CALLPATH
#include "pnmpimod.h"
#include "callpath_module.h"
#endif

extern "C" int handleNewLocation (
        MustParallelId pId,
        MustLocationId lId,
        char *callName,
        int callNameLen
#ifdef USE_CALLPATH
        ,
        int numStackLevels,
        int stackInfosLength, /*Total length of char array for all entries of stack levels*/
        int indicesLength, /*=numStackLevels*3; used for convenience*/
        int* infoIndices, /*Array of size 3*numStackLevels with indices to the individual information pieces, for each level infoIndices[level*3+0]=symName, infoIndices[level*3+1]=lineOffset, infoIndices[level*3+2]=fileModule*/
        char* stackInfos /*All information pieces of all stack levels concatenated, of length stackInfosLength*/
#endif
    );

/**
 * Information for a location.
 * Currently just the name of the MPI call.
 * In future this may be extended to include
 * information about source line/file and so on.
 */
struct MustLocationInfo
{
	std::string callName; /** Name of the API call for this location. */
#ifdef USE_CALLPATH
	std::list<MustStackLevelInfo> stack;
#endif
};

#ifdef USE_CALLPATH
bool operator<(const MustLocationInfo& a, const MustLocationInfo& b)
{
    //Criteria A: callName
    if (a.callName < b.callName)
        return true;

    if (a.callName != b.callName)
        return false;

    //Criteria B: stack depth
    if (a.stack.size() < b.stack.size())
        return true;

    if (a.stack.size() != b.stack.size())
        return false;

    //Critieria C: stack->symName && lineOffset
    std::list<MustStackLevelInfo>::const_iterator aIter, bIter;
    for (aIter = a.stack.begin(), bIter = b.stack.begin(); aIter != a.stack.end(); aIter++, bIter++)
    {
        if (aIter->symName < bIter->symName)
            return true;
        if (aIter->symName != bIter->symName)
            return false;

        if (aIter->lineOffset < bIter->lineOffset)
                    return true;
        if (aIter->lineOffset != bIter->lineOffset)
            return false;
    }

    //We consider this equal if all stack symbol names and the call name are equal
    //(We do not look at file or offsets or module or line, this should usually be ok)
    return false;
}
#endif

/**
 * Maps location id to its information.
 * Must be changed at a later time, as knowing the location id
 * in advance will not be possible if stacks or source line information
 * is added.
 */
#ifndef USE_CALLPATH
std::map<MustLocationId, MustLocationInfo> gKnownLocations;
#else
std::map<int, std::map<MustLocationInfo, MustLocationId> > gKnownLocations;
#endif

/**
 * Get MPI_COMM_WORLD faked to the application.
 *
 * @return a MPI_Comm, used by the application as MPI_COMM_WORLD.
 */
inline MPI_Comm getFakeCommWorld()
{
    static MPI_Comm fakeCommWorld = MPI_COMM_NULL;
    static int  inited, 
                got_comm=0;

    //get rank of this process
    PMPI_Initialized (&inited);

    if (!inited)
        return fakeCommWorld; //TODO: what id to use in such a case ?

    int err;
    PNMPI_modHandle_t handle;
    PNMPI_Service_descriptor_t service;
    PNMPI_Service_Fct_t fct;

    if(!got_comm)
    {
        //We need to check whether MPI_COMM_WORLD was splited
#ifdef PNMPI_FIXED
        err = PNMPI_Service_GetModuleByName("split_processes", &handle);
#else
        char string[512];
        sprintf (string, "%s","split_processes");
        err = PNMPI_Service_GetModuleByName(string, &handle);
#endif
        if (err == PNMPI_SUCCESS)
        {
            err = PNMPI_Service_GetServiceByName(handle, "SplitMod_getMySetComm", "p", &service);
            assert (err == PNMPI_SUCCESS);
            ((int(*)(void*)) service.fct) (&fakeCommWorld);

        }
        else
        {
            //No splitting is active, use MPI_COMM_WORLD
            fakeCommWorld = MPI_COMM_WORLD;
        }
        got_comm=1;
    }

    return fakeCommWorld;
}



/**
 * Builds the parallel id for an MPI call.
 *
 * Currently the implementation uses all bits for the rank.
 * In future this may change to use some of the bits for
 * thread or GPU ids.
 *
 * @return parallel id.
 */
inline MustParallelId buildParallelId ()
{
	static int 	rank = 0,
                inited, 
                got_rank=0;

    //get rank of this process
    PMPI_Initialized (&inited);

    if (!inited)
        return 0; //TODO: what id to use in such a case ?

    if(!got_rank){
        PMPI_Comm_rank (getFakeCommWorld(), &rank);
        got_rank=1;
    }

	return rank;
}

/**
 * Helper function to call the handleNewLocation call when we use the callpath module
 */
#ifdef USE_CALLPATH
/* IMPORTANT: This must closely match whats in LocationImpl.cpp, the use of different data structures requires some redundancy unfortunately */
void createHandleNewLocationCall (MustLocationId id, char* callName, MustLocationInfo &location)
{
       char totalInfo[MUST_MAX_TOTAL_INFO_SIZE];
       int InfoIndices[MUST_MAX_NUM_STACKLEVELS*3];
       int maxtotalLen = MUST_MAX_TOTAL_INFO_SIZE -  MUST_MAX_NUM_STACKLEVELS*4;
       int totalLength = 0;
       int infoIndicesIndex = 0;

       std::list<MustStackLevelInfo>::iterator iter;
       for (iter = location.stack.begin(); iter != location.stack.end(); iter++)
       {
           for (int piece = 0; piece < 3; piece++)
           {
               const char* info = NULL;

               switch (piece)
               {
               case 0: info = iter->symName.c_str();
                   break;
               case 1: info = iter->fileModule.c_str();
                   break;
               case 2: info = iter->lineOffset.c_str();
                   break;
               }

               int i = 0;
               while (info && info[i] != '\0' && totalLength < maxtotalLen)
               {
                   totalInfo[totalLength] = info[i];
                   i++;
                   totalLength++;
               }
               totalInfo[totalLength] = '\0';
               totalLength++;

               InfoIndices[infoIndicesIndex] = totalLength-1;
               infoIndicesIndex++;
           }
       }

       handleNewLocation (
               buildParallelId(),
               id,
               callName,
               location.callName.length()+1,
               location.stack.size(), /*Num stack levels*/
               totalLength, /*stack infos total length*/
               infoIndicesIndex+1, /*indicesLength*/
               InfoIndices, /*infoIndices*/
               totalInfo); /*StackInfos*/
}
#endif

/**
 * Builds the location for the given call name and returns its
 * location id.
 *
 * As we currently only include the call name, we can directly
 * use the given callId as location id.
 * Once we move towards adding stack or source file information
 * we need to add a mapping from the callId and other information
 * to the actual location id.
 *
 * @param callName name of the API call.
 * @param callId unique identifier for this call, comes from the weaver
 * @return location id.
 */
inline MustLocationId buildLocation (const char* callName, int callId)
{
#ifdef USE_CALLPATH
    //a) For the callpath case: get the service function from the callpath module
    static PNMPIMOD_Callpath_GetCallpath_t fct;
    static bool isInitialized = false;

   if (!isInitialized)
   {
        isInitialized = true;

        // find the callpath module
        PNMPI_modHandle_t module;
        int err = PNMPI_Service_GetModuleByName(PNMPI_MODULE_CALLPATH, &module);
        if (err!=PNMPI_SUCCESS) {
            std::cerr << "Couldn't find module " PNMPI_MODULE_CALLPATH << std::endl;
            return err;
        }

        // get the service that will get us the actual call trace
        PNMPI_Service_descriptor_t service;
        err = PNMPI_Service_GetServiceByName(module, PNMPIMOD_Callpath_GetCallpath, "r", &service);
        if (err != PNMPI_SUCCESS) {
            std::cerr << "Couldn't find " PNMPIMOD_Callpath_GetCallpath " service!" << std::endl;
            return err;
        }

        fct = reinterpret_cast<PNMPIMOD_Callpath_GetCallpath_t>(service.fct);
   }

   //b) Build the information for this location (callName + stack)
   std::list<StackInfo> sList = (*fct) ();
   MustLocationInfo thisLocation;
   thisLocation.callName = callName;

   std::list<StackInfo>::iterator sIter;
   for(sIter = sList.begin(); sIter != sList.end(); sIter++)
   {
       MustStackLevelInfo levelInfo;
       levelInfo.symName = sIter->symName;
       levelInfo.fileModule = sIter->fileModule;
       levelInfo.lineOffset = sIter->lineOffset;
       thisLocation.stack.push_back(levelInfo);
   }

   //c) Search in the known locations
   static int nextLocationId = 0;
   MustLocationId id; //return value
   std::map<int, std::map<MustLocationInfo, MustLocationId> >::iterator pos;

   pos = gKnownLocations.find (callId);
   if (pos == gKnownLocations.end ())
   {
       //c-1) Its a new location
       std::map<MustLocationInfo, MustLocationId> temp;
       temp.insert (std::make_pair(thisLocation, nextLocationId));
       gKnownLocations.insert (std::make_pair (callId, temp));
       createHandleNewLocationCall (nextLocationId, (char*)(void*)callName, thisLocation);
       nextLocationId++;
       id = nextLocationId-1;
       }
   else
   {
       //c-2) We have used this call id already (either new or old)
       std::map<MustLocationInfo, MustLocationId>::iterator callIdPos;

       callIdPos = pos->second.find (thisLocation);

       if (callIdPos == pos->second.end())
       {
           //A new stack
           pos->second.insert (std::make_pair(thisLocation, nextLocationId));
           createHandleNewLocationCall (nextLocationId, (char*)(void*)callName, thisLocation);
           nextLocationId++;
           id = nextLocationId-1;
       }
       else
       {
           //A known stack
           id = callIdPos->second;
       }
   }
#else
   //Search in the known locations
   MustLocationId id = callId;

   std::map<MustLocationId, MustLocationInfo>::iterator pos;
   pos = gKnownLocations.find (callId);
   if (pos == gKnownLocations.end ())
   {
       //Its a new location
       MustLocationInfo info;
       info.callName = callName;
       gKnownLocations.insert (std::make_pair (id, info));

       //TODO: can we simply call the wrapper function here ?
       // Note: we cast the const away from the callName, but the function should only have reading access
       handleNewLocation (buildParallelId(), id, (char*)(void*)callName, info.callName.length()+1);
   }
#endif

    return id;
}
